#include "TileContainer.h"
#include "..\common\tinyxml.h"

void CTileContainer::Reset()
{
	m_tilecount = 0;
	m_tiles.clear();
	for (int i=0; i<16; i++) 
	{
		m_listsForCon[i].clear();
	}
	m_tileWidth = 0;
	m_tileHeight = 0;
}

int CTileContainer::ReadXML(string fullpath)
{
	TiXmlDocument tilesetXml(fullpath.c_str());
	if (!(tilesetXml.LoadFile())) 
		return LEVGEN_FILEOPENERROR;

	TiXmlNode* root = tilesetXml.FirstChild(LEVGEN_TILESET_TAG_TILEDEF);
	if (!root) 
		return LEVGEN_INVALIDTILESET;

	TiXmlElement* tilexml = root->FirstChildElement(LEVGEN_TILESET_TAG_TILE);
	if (!tilexml) 
		return LEVGEN_INVALIDTILESET;

	while(tilexml)
	{
		Tile tile;
		tile.isSokobanable = false;
		tile.sokobanBlocks.clear();

		TiXmlElement* imagename = tilexml->FirstChildElement(LEVGEN_TILESET_TAG_IMAGENAME);
		TiXmlElement* typematrix = tilexml->FirstChildElement(LEVGEN_TILESET_TAG_TYPEMATRIX);
		TiXmlElement* sokobanable = tilexml->FirstChildElement(LEVGEN_TILESET_TAG_SOKOBANABLE);
		if (!imagename || !typematrix) 
			return LEVGEN_INVALIDTILESET;


		tile.id = tilexml->Attribute(LEVGEN_TILESET_ATTR_ID);
		if (m_tileWidth == 0) 
		{
			m_tileWidth = atoi(tilexml->Attribute(LEVGEN_TILESET_ATTR_WIDTH));
			m_tileHeight = atoi(tilexml->Attribute(LEVGEN_TILESET_ATTR_HEIGHT));
		}
		tile.width = m_tileWidth;
		tile.height = m_tileHeight;

		tile.imageName = imagename->FirstChild()->Value();

		TiXmlNode* typematrixRow = typematrix->FirstChild();
		for(int i = 0; i< m_tileHeight; i++) 
		{
			string	row = typematrixRow->Value();
			int		lastpos = 0;
			int		count = 0;
			for (int j = 0; j < (int) (row.size()); j++) 
			{
				if (row[j] == ' ')
				{
					if (j - lastpos > 0)
					{
						tile.typeMatrix[i][count++] = atoi(row.substr(lastpos, j - lastpos).c_str());
						if (count >= m_tileWidth) break;
					}
					lastpos = j + 1;
				}
			}
			typematrixRow = typematrixRow->NextSibling();
		}

		if (sokobanable)
		{
			if (1 == atoi(sokobanable->Attribute(LEVGEN_TILESET_ATTR_IS)))
			{
				tile.isSokobanable = true;
			}
			else
			{
				tile.isSokobanable = false;
			}

			if (tile.isSokobanable)
			{
				TiXmlElement* sokobanmatrix = tilexml->FirstChildElement(LEVGEN_TILESET_TAG_SOKOBANMATRIX);
				TiXmlNode* matrixRow = sokobanmatrix->FirstChild();
				for(int i = 0; i< m_tileHeight; i++) 
				{
					string	row = matrixRow->Value();
					int		lastpos = 0;
					int		count = 0;
					for (int j = 0; j < (int) (row.size()); j++) 
					{
						if (row[j] == ' ')
						{
							if (j - lastpos > 0)
							{
								if (atoi(row.substr(lastpos, j - lastpos).c_str()) > 0)
								{
									Coor coor;
									coor.x = count;
									coor.y = i;
									tile.sokobanBlocks.push_back(coor);
								}
								count++;
								if (count >= m_tileWidth) break;
							}
							lastpos = j + 1;
						}
					}
					matrixRow = matrixRow->NextSibling();
				}
			}
		}

		m_tiles.push_back(tile);
		tilexml = tilexml->NextSiblingElement(LEVGEN_TILESET_TAG_TILE);
	}
	

	for (int i = 0; i < (int) (m_tiles.size()); i++) 
	{
		AnalyzeTile(i);
	}

	return LEVGEN_NOERROR;

}

void CTileContainer::AnalyzeTile(int index)
{
	Tile& tile = m_tiles[index];

	memset((void*) tile.isConnected, 0, sizeof(tile.isConnected));
	for (int i = 0; i < NDIRECTION; i++) 
	{
		Direction::Type dir = DIRECTIONS[i];
		int x;
		int y;
		GetGate(dir, &x, &y);
		if (tile.typeMatrix[y][x] > 0) 
			tile.isConnected[dir] = true;
	}

	m_listsForConNoCross[GetIndexForCon(tile.isConnected[Direction::North],
										tile.isConnected[Direction::West],
										tile.isConnected[Direction::East],
										tile.isConnected[Direction::South])].push_back(index);

	memcpy((void*) m_matrixForCrossCheck, (const void*) tile.typeMatrix, sizeof(tile.typeMatrix));

	bool hasAnyConnection = false;
	for (int i = 0; i < NDIRECTION; i++) 
	{
		Direction::Type dir = DIRECTIONS[i];
		if (tile.isConnected[dir] == true) 
		{
			hasAnyConnection = true;
			int x;
			int y;
			GetGate(dir, &x, &y);
			ADFS(x, y);
			break;
		}
	}
	if (!hasAnyConnection)
		return;

	bool flag = true;
	for (int i = 0; i < NDIRECTION; i++) 
	{
		Direction::Type dir = DIRECTIONS[i];
		int x;
		int y;
		GetGate(dir, &x, &y);
		if (tile.isConnected[dir] == true && m_matrixForCrossCheck[y][x] > 0)
		{
			flag = false;
			break;
		}
	}

	if (flag)
	{
		m_listsForCon[GetIndexForCon(tile.isConnected[Direction::North],
										tile.isConnected[Direction::West],
										tile.isConnected[Direction::East],
										tile.isConnected[Direction::South])].push_back(index);
	}
}

void CTileContainer::ADFS(int x, int y)
{
	m_matrixForCrossCheck[y][x] = 0;

	for (int i = 0; i < NDIRECTION; i++) 
	{
		Direction::Type dir = DIRECTIONS[i];

		int nextX = x+DELTAX[dir];
		int nextY = y+DELTAY[dir];
		if (nextX>=0 && nextX<m_tileWidth && nextY>=0 && nextY<m_tileHeight && m_matrixForCrossCheck[nextY][nextX] > 0)
		{
			ADFS(nextX, nextY);
		}
	}
}

int CTileContainer::GetIndexForCon(bool bNorth, bool bWest, bool bEast, bool bSouth)
{
	return (bNorth?1:0) * 8 + (bWest?1:0) * 4 + (bEast?1:0) * 2 + (bSouth?1:0) * 1;
}

void CTileContainer::GetGate(Direction::Type dir, int *x, int *y)
{
	int midWidth = m_tileWidth / 2;
	int midHeight = m_tileHeight / 2;

	if (dir == Direction::North)
	{
		*x = midWidth;
		*y = 0;
	}
	else if (dir == Direction::West)
	{
		*x = 0;
		*y = midHeight;
	}
	else if (dir == Direction::East)
	{
		*x = m_tileWidth-1;
		*y = midHeight;
	}
	else if (dir == Direction::South)
	{
		*x = midWidth;
		*y = m_tileHeight-1;
	}
}
